@@ -1,5 +1,5 @@
 from copy import deepcopy
-from typing import Dict, List, Union
+from typing import Any, Collection, Dict, List, Optional, Union, cast
 
 from cognite.client._api_client import APIClient
 from cognite.client.data_classes import Annotation, AnnotationFilter, AnnotationList, AnnotationUpdate
@@ -9,9 +9,8 @@ from cognite.client.utils._auxiliary import assert_type, to_camel_case
 
 class AnnotationsAPI(APIClient):
     _RESOURCE_PATH = "/annotations"
-    _LIST_CLASS = AnnotationList
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
         super().__init__(*args, **kwargs)
 
     def create(self, annotations: Union[Annotation, List[Annotation]]) -> Union[Annotation, AnnotationList]:
@@ -24,7 +23,9 @@ class AnnotationsAPI(APIClient):
             Union[Annotation, AnnotationList]: created annotation(s)
         """
         assert_type(annotations, "annotations", [Annotation, list])
-        return self._create_multiple(resource_path=self._RESOURCE_PATH + "/", items=annotations)
+        return self._create_multiple(
+            list_cls=AnnotationList, resource_cls=Annotation, resource_path=self._RESOURCE_PATH + "/", items=annotations
+        )
 
     def suggest(self, annotations: Union[Annotation, List[Annotation]]) -> Union[Annotation, AnnotationList]:
         """Suggest annotations
@@ -38,14 +39,19 @@ class AnnotationsAPI(APIClient):
         assert_type(annotations, "annotations", [Annotation, list])
         # Deal with status fields in both cases: Single item and list of items
         items = (
-            [AnnotationsAPI._sanitize_suggest_item(ann) for ann in annotations]
+            [self._sanitize_suggest_item(ann) for ann in annotations]
             if isinstance(annotations, list)
-            else AnnotationsAPI._sanitize_suggest_item(annotations)
+            else self._sanitize_suggest_item(annotations)
+        )
+        return self._create_multiple(
+            list_cls=AnnotationList,
+            resource_cls=Annotation,
+            resource_path=self._RESOURCE_PATH + "/suggest",
+            items=items,  # type: ignore[arg-type]
         )
-        return self._create_multiple(resource_path=self._RESOURCE_PATH + "/suggest", items=items)
 
     @staticmethod
-    def _sanitize_suggest_item(annotation: Annotation) -> Dict[str, any]:
+    def _sanitize_suggest_item(annotation: Union[Annotation, Dict[str, Any]]) -> Dict[str, Any]:
         # Check that status is set to suggested if it is set and afterwards remove it
         item = annotation.dump(camel_case=True) if isinstance(annotation, Annotation) else deepcopy(annotation)
         if "status" in item:
@@ -78,13 +84,17 @@ class AnnotationsAPI(APIClient):
                 {to_camel_case(k): v for k, v in f.items()} for f in filter["annotatedResourceIds"]
             ]
 
-        return self._list(method="POST", limit=limit, filter=filter)
+        return self._list(list_cls=AnnotationList, resource_cls=Annotation, method="POST", limit=limit, filter=filter)
 
     @staticmethod
-    def _convert_resource_to_patch_object(resource: CogniteResource, update_attributes: List[str]):
+    def _convert_resource_to_patch_object(
+        resource: CogniteResource, update_attributes: Collection[str]
+    ) -> Dict[str, Dict[str, Dict]]:
         if not isinstance(resource, Annotation):
             return APIClient._convert_resource_to_patch_object(resource, update_attributes)
         annotation: Annotation = resource
+
+        assert annotation.id is not None
         annotation_update = AnnotationUpdate(id=annotation.id)
         for attr in update_attributes:
             getattr(annotation_update, attr).set(getattr(annotation, attr))
@@ -98,7 +108,9 @@ class AnnotationsAPI(APIClient):
         Args:
             id (Union[int, List[int]]): ID or list of IDs to be deleted
         """
-        return self._update_multiple(items=item)
+        return self._update_multiple(
+            list_cls=AnnotationList, resource_cls=Annotation, update_cls=AnnotationUpdate, items=item
+        )
 
     def delete(self, id: Union[int, List[int]]) -> None:
         """Delete annotations
@@ -118,9 +130,12 @@ class AnnotationsAPI(APIClient):
             AnnotationList: list of annotations
         """
         assert_type(ids, "ids", [List], allow_none=False)
-        return self._retrieve_multiple(ids=ids, wrap_ids=True)
+        return cast(
+            AnnotationList,
+            self._retrieve_multiple(list_cls=AnnotationList, resource_cls=Annotation, ids=ids, wrap_ids=True),
+        )
 
-    def retrieve(self, id: int) -> Annotation:
+    def retrieve(self, id: int) -> Optional[Annotation]:
         """Retrieve an annotation by id
 
         Args:
@@ -130,4 +145,7 @@ class AnnotationsAPI(APIClient):
             Annotation: annotation requested
         """
         assert_type(id, "id", [int], allow_none=False)
-        return self._retrieve_multiple(ids=id, wrap_ids=True)
+        return cast(
+            Optional[Annotation],
+            self._retrieve_multiple(list_cls=AnnotationList, resource_cls=Annotation, ids=id, wrap_ids=True),
+        )
