@@ -1,4 +1,4 @@
-from typing import Generator, List, Union
+from typing import Any, Dict, Iterator, List, Union, cast
 
 from cognite.client._api_client import APIClient
 from cognite.client.data_classes import LabelDefinition, LabelDefinitionFilter, LabelDefinitionList
@@ -6,9 +6,8 @@ from cognite.client.data_classes import LabelDefinition, LabelDefinitionFilter,
 
 class LabelsAPI(APIClient):
     _RESOURCE_PATH = "/labels"
-    _LIST_CLASS = LabelDefinitionList
 
-    def __iter__(self) -> Generator[LabelDefinition, None, None]:
+    def __iter__(self) -> Iterator[LabelDefinition]:
         """Iterate over Labels
 
         Fetches Labels as they are iterated over, so you keep a limited number of Labels in memory.
@@ -16,17 +15,48 @@ class LabelsAPI(APIClient):
         Yields:
             Type: yields Labels one by one.
         """
-        return self.__call__()
+        return cast(Iterator[LabelDefinition], self.__call__())
 
-    def __call__(self, name: str = None, external_id_prefix: str = None, limit: int = None, chunk_size: int = None):
-        filter = LabelDefinitionFilter(name=name, external_id_prefix=external_id_prefix).dump(camel_case=True)
-        return self._list_generator(method="POST", limit=limit, filter=filter, chunk_size=chunk_size)
+    def __call__(
+        self,
+        name: str = None,
+        external_id_prefix: str = None,
+        limit: int = None,
+        chunk_size: int = None,
+        data_set_ids: List[int] = None,
+        data_set_external_ids: List[str] = None,
+    ) -> Union[Iterator[LabelDefinition], Iterator[LabelDefinitionList]]:
+        data_set_ids_processed = None
+        if data_set_ids or data_set_external_ids:
+            data_set_ids_processed = cast(
+                List[Dict[str, Any]], self._process_ids(data_set_ids, data_set_external_ids, wrap_ids=True)
+            )
+        filter = LabelDefinitionFilter(
+            name=name, external_id_prefix=external_id_prefix, data_set_ids=data_set_ids_processed
+        ).dump(camel_case=True)
+        return self._list_generator(
+            list_cls=LabelDefinitionList,
+            resource_cls=LabelDefinition,
+            method="POST",
+            limit=limit,
+            filter=filter,
+            chunk_size=chunk_size,
+        )
 
-    def list(self, name: str = None, external_id_prefix: str = None, limit: int = 25) -> LabelDefinitionList:
+    def list(
+        self,
+        name: str = None,
+        external_id_prefix: str = None,
+        data_set_ids: List[int] = None,
+        data_set_external_ids: List[str] = None,
+        limit: int = 25,
+    ) -> LabelDefinitionList:
         """`List Labels <https://docs.cognite.com/api/v1/#operation/listLabels>`_
 
         Args:
             name (str): returns the label definitions matching that name
+            data_set_ids (List[int]): return only labels in the data sets with these ids.
+            data_set_external_ids (List[str]): return only labels in the data sets with these external ids.
             external_id_prefix (str): filter label definitions with external ids starting with the prefix specified
             limit (int, optional): Maximum number of label definitions to return.
 
@@ -55,8 +85,17 @@ class LabelsAPI(APIClient):
                 >>> for label_list in c.labels(chunk_size=2500):
                 ...     label_list # do something with the type definitions
         """
-        filter = LabelDefinitionFilter(name=name, external_id_prefix=external_id_prefix).dump(camel_case=True)
-        return self._list(method="POST", limit=limit, filter=filter)
+        data_set_ids_processed = None
+        if data_set_ids or data_set_external_ids:
+            data_set_ids_processed = cast(
+                List[Dict[str, Any]], self._process_ids(data_set_ids, data_set_external_ids, wrap_ids=True)
+            )
+        filter = LabelDefinitionFilter(
+            name=name, external_id_prefix=external_id_prefix, data_set_ids=data_set_ids_processed
+        ).dump(camel_case=True)
+        return self._list(
+            list_cls=LabelDefinitionList, resource_cls=LabelDefinition, method="POST", limit=limit, filter=filter
+        )
 
     def create(
         self, label: Union[LabelDefinition, List[LabelDefinition]]
@@ -84,7 +123,7 @@ class LabelsAPI(APIClient):
                 raise TypeError("'label' must be of type List[LabelDefinition]")
         elif not isinstance(label, LabelDefinition):
             raise TypeError("'label' must be of type LabelDefinition")
-        return self._create_multiple(items=label)
+        return self._create_multiple(list_cls=LabelDefinitionList, resource_cls=LabelDefinition, items=label)
 
     def delete(self, external_id: Union[str, List[str]] = None) -> None:
         """`Delete one or more label definitions <https://docs.cognite.com/api/v1/#operation/deleteLabels>`_
