@@ -1,11 +1,25 @@
-from typing import List, Union
+from typing import Any, Dict, List, Optional, Union, cast
 
+from cognite.client import utils
 from cognite.client._api_client import APIClient
-from cognite.client.data_classes.templates import *
+from cognite.client.data_classes.templates import (
+    GraphQlResponse,
+    TemplateGroup,
+    TemplateGroupList,
+    TemplateGroupVersion,
+    TemplateGroupVersionList,
+    TemplateInstance,
+    TemplateInstanceList,
+    TemplateInstanceUpdate,
+    View,
+    ViewList,
+    ViewResolveItem,
+    ViewResolveList,
+)
 
 
 class TemplatesAPI(APIClient):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
         super().__init__(*args, **kwargs)
         self.groups = TemplateGroupsAPI(*args, **kwargs)
         self.versions = TemplateGroupVersionsAPI(*args, **kwargs)
@@ -57,9 +71,8 @@ class TemplatesAPI(APIClient):
 
 class TemplateGroupsAPI(APIClient):
     _RESOURCE_PATH = "/templategroups"
-    _LIST_CLASS = TemplateGroupList
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
         super().__init__(*args, **kwargs)
 
     def create(
@@ -83,7 +96,7 @@ class TemplateGroupsAPI(APIClient):
                 >>> template_group_2 = TemplateGroup("sdk-test-group-2", "This is another test group")
                 >>> c.templates.groups.create([template_group_1, template_group_2])
         """
-        return self._create_multiple(items=template_groups)
+        return self._create_multiple(list_cls=TemplateGroupList, resource_cls=TemplateGroup, items=template_groups)
 
     def upsert(
         self, template_groups: Union[TemplateGroup, List[TemplateGroup]]
@@ -110,9 +123,13 @@ class TemplateGroupsAPI(APIClient):
         path = self._RESOURCE_PATH + "/upsert"
         is_single = not isinstance(template_groups, list)
         if is_single:
-            template_groups = [template_groups]
-        updated = self._post(path, {"items": [item.dump(camel_case=True) for item in template_groups]}).json()["items"]
-        res = self._LIST_CLASS._load(updated, cognite_client=self._cognite_client)
+            template_groups_processed: List[TemplateGroup] = cast(List[TemplateGroup], [template_groups])
+        else:
+            template_groups_processed = cast(List[TemplateGroup], template_groups)
+        updated = self._post(
+            path, {"items": [item.dump(camel_case=True) for item in template_groups_processed]}
+        ).json()["items"]
+        res = TemplateGroupList._load(updated, cognite_client=self._cognite_client)
         if is_single:
             return res[0]
         return res
@@ -134,7 +151,16 @@ class TemplateGroupsAPI(APIClient):
                 >>> c = CogniteClient()
                 >>> res = c.templates.groups.retrieve_multiple(external_ids=["abc", "def"])
         """
-        return self._retrieve_multiple(external_ids=external_ids, ignore_unknown_ids=ignore_unknown_ids, wrap_ids=True)
+        return cast(
+            TemplateGroupList,
+            self._retrieve_multiple(
+                list_cls=TemplateGroupList,
+                resource_cls=TemplateGroup,
+                external_ids=external_ids,
+                ignore_unknown_ids=ignore_unknown_ids,
+                wrap_ids=True,
+            ),
+        )
 
     def list(self, limit: int = 25, owners: List[str] = None) -> TemplateGroupList:
         """`Lists template groups stored in the project based on a query filter given in the payload of this request.`
@@ -158,11 +184,17 @@ class TemplateGroupsAPI(APIClient):
         filter = {}
         if owners is not None:
             filter["owners"] = owners
-        return self._list(method="POST", limit=limit, filter=filter, partitions=None, sort=None)
+        return self._list(
+            list_cls=TemplateGroupList,
+            resource_cls=TemplateGroup,
+            method="POST",
+            limit=limit,
+            filter=filter,
+            partitions=None,
+            sort=None,
+        )
 
-    def delete(
-        self, external_ids: Union[str, List[str]], ignore_unknown_ids: bool = False
-    ) -> Union[TemplateGroup, TemplateGroupList]:
+    def delete(self, external_ids: Union[str, List[str]], ignore_unknown_ids: bool = False) -> None:
         """`Delete one or more template groups.`
 
         Args:
@@ -177,21 +209,20 @@ class TemplateGroupsAPI(APIClient):
 
                 >>> from cognite.client import CogniteClient
                 >>> c = CogniteClient()
-                >>> c.templates.groups.delete(external_id=["a", "b"])
+                >>> c.templates.groups.delete(external_ids=["a", "b"])
         """
         return self._delete_multiple(
-            True, external_ids=external_ids, extra_body_fields={"ignoreUnknownIds": ignore_unknown_ids}
+            wrap_ids=True, external_ids=external_ids, extra_body_fields={"ignoreUnknownIds": ignore_unknown_ids}
         )
 
 
 class TemplateGroupVersionsAPI(APIClient):
     _RESOURCE_PATH = "/templategroups/{}/versions"
-    _LIST_CLASS = TemplateGroupVersionList
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any) -> None:
         super().__init__(*args, **kwargs)
 
-    def upsert(self, external_id: str, version: TemplateGroupVersion):
+    def upsert(self, external_id: str, version: TemplateGroupVersion) -> TemplateGroupVersion:
         """`Upsert a template group version.`
         A Template Group update supports specifying different conflict modes, which is used when an existing schema already exists.
 
@@ -232,8 +263,8 @@ class TemplateGroupVersionsAPI(APIClient):
                 >>> c.templates.versions.upsert(template_group.external_id, template_group_version)
         """
         resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id) + "/upsert"
-        version = self._post(resource_path, version.dump(camel_case=True)).json()
-        return TemplateGroupVersion._load(version)
+        version_res = self._post(resource_path, version.dump(camel_case=True)).json()
+        return TemplateGroupVersion._load(version_res)
 
     def list(
         self, external_id: str, limit: int = 25, min_version: Optional[int] = None, max_version: Optional[int] = None
@@ -264,7 +295,14 @@ class TemplateGroupVersionsAPI(APIClient):
             filter["minVersion"] = min_version
         if max_version is not None:
             filter["maxVersion"] = max_version
-        return self._list(resource_path=resource_path, method="POST", limit=limit, filter=filter)
+        return self._list(
+            list_cls=TemplateGroupVersionList,
+            resource_cls=TemplateGroupVersion,
+            resource_path=resource_path,
+            method="POST",
+            limit=limit,
+            filter=filter,
+        )
 
     def delete(self, external_id: str, version: int) -> None:
         """`Delete a template group version.`
@@ -329,10 +367,12 @@ class TemplateInstancesAPI(APIClient):
                 >>> c.templates.instances.create("sdk-test-group", 1, [template_instance_1, template_instance_2])
         """
         resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
-        return self._create_multiple(resource_path=resource_path, items=instances)
+        return self._create_multiple(
+            list_cls=TemplateInstanceList, resource_cls=TemplateInstance, resource_path=resource_path, items=instances
+        )
 
     def upsert(
-        self, external_id: str, version: int, instances: Union[TemplateGroup, List[TemplateGroup]]
+        self, external_id: str, version: int, instances: Union[TemplateInstance, List[TemplateInstance]]
     ) -> Union[TemplateInstance, TemplateInstanceList]:
         """`Upsert one or more template instances.`
         Will overwrite existing instances.
@@ -384,7 +424,7 @@ class TemplateInstancesAPI(APIClient):
 
     def update(
         self, external_id: str, version: int, item: Union[TemplateInstanceUpdate, List[TemplateInstanceUpdate]]
-    ) -> Union[TemplateInstance, List[TemplateInstance]]:
+    ) -> Union[TemplateInstance, TemplateInstanceList]:
         """`Update one or more template instances`
         Args:
             item (Union[TemplateInstanceUpdate, List[TemplateInstanceUpdate]]): Templates instance(s) to update
@@ -402,7 +442,13 @@ class TemplateInstancesAPI(APIClient):
                 >>> res = c.templates.instances.update("sdk-test-group", 1, my_update)
         """
         resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
-        return self._update_multiple(items=item, resource_path=resource_path)
+        return self._update_multiple(
+            list_cls=TemplateInstanceList,
+            resource_cls=TemplateInstance,
+            update_cls=TemplateInstanceUpdate,
+            items=item,
+            resource_path=resource_path,
+        )
 
     def retrieve_multiple(
         self, external_id: str, version: int, external_ids: List[str], ignore_unknown_ids: bool = False
@@ -426,8 +472,16 @@ class TemplateInstancesAPI(APIClient):
                 >>> res = c.templates.instances.retrieve_multiple(external_id="sdk-test-group", version=1, external_ids=["abc", "def"])
         """
         resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
-        return self._retrieve_multiple(
-            resource_path=resource_path, external_ids=external_ids, ignore_unknown_ids=ignore_unknown_ids, wrap_ids=True
+        return cast(
+            TemplateInstanceList,
+            self._retrieve_multiple(
+                list_cls=TemplateInstanceList,
+                resource_cls=TemplateInstance,
+                resource_path=resource_path,
+                external_ids=external_ids,
+                ignore_unknown_ids=ignore_unknown_ids,
+                wrap_ids=True,
+            ),
         )
 
     def list(
@@ -460,12 +514,19 @@ class TemplateInstancesAPI(APIClient):
                 >>> template_instances_list = c.templates.instances.list("template-group-ext-id", 1, limit=5)
         """
         resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
-        filter = {}
+        filter: Dict[str, Any] = {}
         if data_set_ids is not None:
             filter["dataSetIds"] = data_set_ids
         if template_names is not None:
             filter["templateNames"] = template_names
-        return self._list(resource_path=resource_path, method="POST", limit=limit, filter=filter)
+        return self._list(
+            list_cls=TemplateInstanceList,
+            resource_cls=TemplateInstance,
+            resource_path=resource_path,
+            method="POST",
+            limit=limit,
+            filter=filter,
+        )
 
     def delete(self, external_id: str, version: int, external_ids: List[str], ignore_unknown_ids: bool = False) -> None:
         """`Delete one or more template instances.`
@@ -497,7 +558,6 @@ class TemplateInstancesAPI(APIClient):
 
 class TemplateViewsAPI(APIClient):
     _RESOURCE_PATH = "/templategroups/{}/versions/{}/views"
-    _LIST_CLASS = ViewList
 
     def create(self, external_id: str, version: int, views: Union[View, List[View]]) -> Union[View, ViewList]:
         """`Create one or more template views.`
@@ -514,7 +574,7 @@ class TemplateViewsAPI(APIClient):
             Create new views:
 
                 >>> from cognite.client import CogniteClient
-                >>> from cognite.client.data_classes import View
+                >>> from cognite.client.data_classes.templates import View
                 >>> c = CogniteClient()
                 >>> view = View(external_id="view",
                 >>>             source=Source(
@@ -533,7 +593,7 @@ class TemplateViewsAPI(APIClient):
                 >>> c.templates.views.create("sdk-test-group", 1, [view])
         """
         resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
-        return self._create_multiple(resource_path=resource_path, items=views)
+        return self._create_multiple(list_cls=ViewList, resource_cls=View, resource_path=resource_path, items=views)
 
     def upsert(self, external_id: str, version: int, views: Union[View, List[View]]) -> Union[View, ViewList]:
         """`Upsert one or more template views.`
@@ -550,7 +610,7 @@ class TemplateViewsAPI(APIClient):
             Upsert new views:
 
                 >>> from cognite.client import CogniteClient
-                >>> from cognite.client.data_classes import View
+                >>> from cognite.client.data_classes.templates import View
                 >>> c = CogniteClient()
                 >>> view = View(external_id="view",
                 >>>             source=Source(
@@ -580,7 +640,7 @@ class TemplateViewsAPI(APIClient):
         return res
 
     def resolve(
-        self, external_id: str, version: int, view_external_id: str, input: Optional[Dict[str, any]], limit: int = 25
+        self, external_id: str, version: int, view_external_id: str, input: Optional[Dict[str, Any]], limit: int = 25
     ) -> ViewResolveList:
         """`Resolves a View.`
         It resolves the source specified in a View with the provided input and applies the mapping rules to the response.
@@ -604,9 +664,10 @@ class TemplateViewsAPI(APIClient):
         """
         url_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version) + "/resolve"
         return self._list(
+            list_cls=ViewResolveList,
+            resource_cls=ViewResolveItem,
             url_path=url_path,
             method="POST",
-            cls=ViewResolveList,
             limit=limit,
             other_params={"externalId": view_external_id, "input": input},
         )
@@ -632,7 +693,7 @@ class TemplateViewsAPI(APIClient):
                 >>> c.templates.views.list("template-group-ext-id", 1, limit=5)
         """
         resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
-        return self._list(resource_path=resource_path, method="POST", limit=limit)
+        return self._list(list_cls=ViewList, resource_cls=View, resource_path=resource_path, method="POST", limit=limit)
 
     def delete(
         self, external_id: str, version: int, view_external_id: Union[List[str], str], ignore_unknown_ids: bool = False
